/***********************************************************
This program is implemented by Chris Chartier, Kenneth King, 
and Andrei Ostapenko. Please do not distribute or re-use the
code without their permission.
***********************************************************/

#include "puzzle.h"

/* a few globals */

const char EMPTY = ASCIISTART;
int iNewDepth = 0;
PuzNode *Goal = NULL;

/* end of globals */

/***************** DEBUG FUNCTIONS */
void PrintHashTable();
void PrintNode(PuzNode *Node);
unsigned long NumberHashElements();
/***************** DEBUG FUNCTIONS */

/**********************Main*************************/
 
int main()
{

   PuzNode *Start =  NULL;
   PuzNode *Solution = NULL;
   
   char textName[65];
   int iReturn = -1;
   int iDepth = 0;

   printf("\nEnter name of input puzzle file: ");
   scanf("%s",textName);

   Start =  ReadFile (textName);


   if(NULL == Start)
   {
      printf("Invalid File for Start Node.\n");
      return -1;
   }
/* Goal is a global */

   Goal = CreateGoalNode();

   iReturn = CheckIfGoal(Start);
   if (PTRUE == iReturn)
   {
	printf("Start Configuration was the solution.\nSolved in 0 moves.\n");
	return -1;
   }

   iReturn = CanSolve(Start);
   if (PFALSE == iReturn)
   {
	printf("Start Configuration is unsolvable.\n");
	return -1;
   }

   iNewDepth = FindDistance(Start);
   iDepth = UpdateDepth();

   do
   {
	printf("Trying depth %d\n", iDepth);
	/*system("pause");*/
	Start =  ReadFile (textName);
	Solution = DepthFirstSearch(iDepth, Start);	
   
	iDepth = UpdateDepth(); 

   } while (NULL == Solution);

   /*PrintSolution(Solution);*/
   PrintSolutionReverse(Solution);
   return 0;

}

/****************************************************************************/


/*************************************************************************
 * ReadFromFile (char *FileName)
 * returns the a new node which is read from the input file
 *************************************************************************/

/* Utility to read the initial state file */

PuzNode* ReadFile (char* FileName)
{
   FILE* fp = NULL;
   char c;
   char tmpStr[51];
   int i;
   char Offset;
   PuzNode *structNewNode = NULL;

   Offset = ASCIISTART;

   for (i=0; i<51; i++)
      tmpStr [i] = '\0';

/*check for no input */

   if (NULL == FileName)
   {
      return NULL; /* empty string for filename */
   }

/* Try to open file */
   fp = fopen(FileName, "r");

   if (NULL == fp)
   {
      return NULL; /* return NULL if open of file failed */
   }

	structNewNode = (PuzNode *)malloc(sizeof(PuzNode));

	structNewNode->m_iMovesToGetHere = 0;
	structNewNode->m_cMoveThatGotMeHere = NOMOVE;
	structNewNode->m_ptrParent = NULL;

   c = fgetc(fp);
   i = 0;
   while (!feof(fp))
   {
      if (isdigit(c))
      {
         tmpStr[i] = c;
         i++;
      }
      c = fgetc(fp);
   }  

   for (i=0; i<25; i++)
   {
      structNewNode->m_strPuz[i] = 10*(tmpStr[2*i] - '0') + tmpStr[2*i+1] - '0' + Offset;
   }

   structNewNode->m_cHeuristic = FindDistance(structNewNode);

return structNewNode;

}

/*************************************************************************
 * UpdateCost (PuzNode *Node)
 * Computes the f-cost (sum of heuristic and moves-to-get-here for the given
 * given node.
 *************************************************************************/
int UpdateCost (PuzNode *Node)
{
	return (Node->m_cHeuristic + Node->m_iMovesToGetHere);
}

/*************************************************************************
 * CreateGoalNode()
 * Creates the (global) goal node.
 *************************************************************************/

PuzNode * CreateGoalNode()
{
	int iLoop = 0;

	PuzNode *structNewNode = NULL;
	structNewNode = (PuzNode *)malloc(sizeof(PuzNode));

		
	structNewNode->m_iMovesToGetHere = 0;
	structNewNode->m_cMoveThatGotMeHere = NOMOVE;
	structNewNode->m_ptrParent = NULL;
	
/* Add ASCIISTART to each char in string to get in the valid string character range.*/

	for (iLoop = 0; iLoop < 24; iLoop++)
	{
		structNewNode->m_strPuz[iLoop] = iLoop + 1 + ASCIISTART;
	}
	structNewNode->m_strPuz[24] = ASCIISTART;
	structNewNode->m_strPuz[25] = '\0';
	structNewNode->m_cHeuristic = FindDistance(structNewNode);

	return structNewNode;
}

/*************************************************************************
 * FindDistance(PuzNode* node)
 * returns the city block (Manhattan) distance for a puzzle state
 **************************************************************************/

int FindDistance (PuzNode* node)   

{
   char Offset;
   int a, h, row, col;

   Offset = ASCIISTART;
   h = 0;

   for (row=0; row<5; row++)
   {
      for (col=0; col<5; col++)
      {
         a = node->m_strPuz[5*row+col] - Offset - 1;
         if (a != -1)
            h = abs(a/5 - row) + abs(a%5 - col) + h;
      }
   }
   return h;
}

/*************************************************************************
 * int CanSolve (PuzNode * TestNode)
 * returns PTRUE if a puzzle is solvable, PFALSE if not solvable. 
 **************************************************************************/

int CanSolve (PuzNode * TestNode)
{
   int i= 0;
   int j = 0;
   int count = 0;
   int swap = 0;
   int size = 25;
   int node[25];
   
/* initialize node[25] */

   for (i = 0; i < 25; i++)
   {
      node[i] = 99;
   }

/* copy TestNode into node[] */

   for (i = 0; i < size; i++)
   {
      node[i] = TestNode->m_strPuz[i];
   }

/* move blank tile to front */

   for (i = 1; i < size; i++)
   {
      for (j = 0; j < size - 1; j++)
	  {
         if ((ASCIISTART) == node[j+1])
         {
            swap = node[j];
            node[j] = node[j+1];
            node[j+1] = swap;
         }
	  }
   }

/* put puzzle in order using bubble sort */
/* and count number of swaps needed. */

   for (i = 1; i < size; i++)
   {
      for (j = 0; j < size - 1; j++)
	  {
         if (node[j] > node[j+1])
         {
            swap = node[j];
            node[j] = node[j+1];
            node[j+1] = swap;
            count++;
         }
	  }
  }

/* an even number of swaps means that the puzzle is solvable */

   if (0 == (swap = count%2))
      return PTRUE;
   else
      return PFALSE;
}

/*************************************************************************
 * PuzNode * DepthFirstSearch(int DepthLimit, PuzNode *Node)
 * returns NULL if solution not found, solution node if found. 
 **************************************************************************/

PuzNode * DepthFirstSearch(int DepthLimit, PuzNode *Node)
{
      int i,j;
	PuzNode *p = NULL;
	PuzNode *temp;
	PuzNode *List[4] = {NULL, NULL, NULL, NULL};
	
	if (PFALSE == CheckIfGoal(Node))
   {

	   /* Check Limit */

	    if (DepthLimit < Node->m_iMovesToGetHere )
		{
			return NULL;
		}
   /* Expand Node's children */

		List[0] = Transition(Node, DOWN, DepthLimit);
		List[1] = Transition(Node, LEFT, DepthLimit);
		List[2] = Transition(Node, RIGHT, DepthLimit);
		List[3] = Transition(Node, UP, DepthLimit);
		
		if ( (NULL == List[0]) &&
		     (NULL == List[1]) &&
		     (NULL == List[2]) &&
		     (NULL == List[3]) )
		{
			free(Node);
			Node = NULL;
			return NULL;
		}

/* do search using node with lowest heuristic first */

		for (j = 0; j < 4; j++)
		{
			for (i=j; i < 4; i++)
			{
				if ((List[i]!=NULL)&&(List[j]!=NULL) )
				{
					if ((List[i]->m_cHeuristic) < (List[j]->m_cHeuristic))
					{
						temp = List[i];
						List[i] = List[j];
						List[j] = temp;
					}
				}
			}
		}

		if (NULL != List[0])
		{
			p = DepthFirstSearch (DepthLimit,List[0]);

			if (NULL != p)
			{
				return p;
			}
		}

		if (NULL != List[1])
		{
			p = DepthFirstSearch (DepthLimit,List[1]);

			if (NULL != p)
			{
				return p;
			}
		}

		if (NULL != List[2])
		{
			p = DepthFirstSearch (DepthLimit,List[2]);

			if (NULL != p)
			{
				return p;
			}
		}

		if (NULL != List[3])
		{
			p = DepthFirstSearch (DepthLimit,List[3]);
			
			if (NULL != p)
			{
				return p;
			}
		}

		free(Node);
		return NULL;
   }
   else /* solution found !! */
   {
	   return Node;
   }
}

/*************************************************************************
 * void ShowSolution (PuzNode* endNode, PuzNode* startNode)
 * prints move-that-got-me-here between startNode and endNode. 
 **************************************************************************/

void ShowSolution (PuzNode* endNode, PuzNode* startNode)
{
   PuzNode* prevNode;
   prevNode = endNode->m_ptrParent;

   if (!(CompareNodes(endNode, startNode)))
      ShowSolution (prevNode, startNode);

   if (!(CompareNodes(endNode, startNode)))
      printf ("%c\n",endNode->m_cMoveThatGotMeHere);

   return;
}

/*************************************************************************
 * int CompareNodes (PuzNode* nodeA, PuzNode* nodeB)
 * returns PTRUE if nodes are the same, PFALSE if different. 
 **************************************************************************/

int CompareNodes (PuzNode* nodeA, PuzNode* nodeB)
{
   int a;

   a = strcmp (nodeA->m_strPuz, nodeB->m_strPuz);
   if (0 == a)
      return PTRUE;
   else
      return PFALSE;
}

/*************************************************************************
 * int CheckIfGoal (PuzNode* node)
 * compares node to Goal (global). returns PTRUE if same, PFALSE if different. 
 **************************************************************************/

int CheckIfGoal (PuzNode* node)
{
   int a;

/* Goal is global */

   a = strcmp (node->m_strPuz, Goal->m_strPuz);
   if (0 == a)
      return PTRUE;
   else
      return PFALSE;
}

/*************************************************************************
 * int UpdateDepth (void)
 * set depth to newdepth and increase newdepth 
 **************************************************************************/

int UpdateDepth (void)
{
   int depth;

/* iNewDepth is global */

   depth = iNewDepth;
   iNewDepth = depth + STEPSIZE; 
   return depth;
}

/*******************End of Roddy's functions ***************************/



/* -----  Andrei's functions */

PuzNode * Transition(PuzNode *Parent, char MoveDirection, int depth)
{
	PuzNode * pTempNode;
	int a;
	int iEmptyPosition = -1;
	char *strTemp = NULL;

	/* Don't go further if goal is already found */

	if ((PTRUE == CheckIfGoal(Parent)) || (NOMOVE == (int)MoveDirection))
	{
		return NULL;
	}

	/* Don't go further if invalid move direction */

	if ( (UP != (int)MoveDirection) && 
		 (DOWN != (int)MoveDirection) && 
		 (RIGHT != (int)MoveDirection) && 
		 (LEFT != (int)MoveDirection))
	{
		return NULL;
	}

	/* Allocate memory for new node */
	pTempNode = (PuzNode *) malloc(sizeof(PuzNode));

	/* Find empty position (0 - 24); */
	strTemp = strchr(Parent->m_strPuz, ASCIISTART);
	iEmptyPosition = strTemp - Parent->m_strPuz;
	
   	/* copy string to temp node (before trying swap) */
	strcpy(pTempNode->m_strPuz , Parent->m_strPuz);

	/* If its not a legal move - return NULL, otherwise  swap places */
	if ((LEFT == MoveDirection) && 
          ((0 == (iEmptyPosition % 5)) || (RIGHT == Parent->m_cMoveThatGotMeHere)))
	{
		free (pTempNode);
		return NULL;
	}

	if ((RIGHT == MoveDirection) && 
          ((4 == (iEmptyPosition % 5)) || (LEFT == Parent->m_cMoveThatGotMeHere)))
	{
		free (pTempNode);
		return NULL;
	}

	if ((DOWN == MoveDirection) && 
          ((19 < (iEmptyPosition)) || (UP == Parent->m_cMoveThatGotMeHere)))
	{
		free (pTempNode);
		return NULL;
	}

	if ((UP == MoveDirection) && 
          ((5 > (iEmptyPosition)) || (DOWN == Parent->m_cMoveThatGotMeHere)))
	{
		free (pTempNode);
		return NULL;
	}

	/* swap blank tile and move tile */
	pTempNode->m_strPuz[iEmptyPosition] = Parent->m_strPuz[iEmptyPosition + (int)MoveDirection];
	pTempNode->m_strPuz[iEmptyPosition + (int)MoveDirection] = EMPTY;
	
	/* check for duplicate 3 moves ago  (not implemented yet) */

/*    if (pTempNode->m_strPuz == (Parent->m_ptrParent->m_ptrParent->m_ptrParent->m_strPuz))
      {
		free (pTempNode);
		return NULL;
	}
*/
	/* Create new node and populate it with values */
	
	pTempNode->m_iMovesToGetHere = Parent->m_iMovesToGetHere + 1;
	pTempNode->m_cHeuristic = FindDistance(pTempNode);
	pTempNode->m_cMoveThatGotMeHere = MoveDirection;
      pTempNode->m_ptrParent = Parent;

      /* if f-cost is over limit, update iNewDepth and return NULL */

      if ((a = (pTempNode->m_iMovesToGetHere + pTempNode->m_cHeuristic)) > depth)
	{
            if (a < iNewDepth)
            {
               iNewDepth = a;
            }
		free (pTempNode);
		return NULL;
	}

 	return pTempNode;
}


/**************/
/* PrintSolution*/
void PrintSolutionReverse(PuzNode *Solution)
{
	int iArr[200];
	int i = 0;
	printf("Number of moves: %d\n", Solution->m_iMovesToGetHere);

	iArr[0] = Solution->m_cMoveThatGotMeHere;
	while (NULL != Solution->m_ptrParent)
	{
		i++;
		iArr[i] = Solution->m_ptrParent->m_cMoveThatGotMeHere;
		Solution = Solution->m_ptrParent;
	}

	for (; 0 <= i; i--)
	{
			switch (iArr[i])
			{
			case DOWN:
			{
				printf("DOWN\n", i);
				break;
			}
			case LEFT:
			{
				printf("LEFT\n", i);
				break;
			}
			case RIGHT:
			{
				printf("RIGHT\n");
				break;
			}
			case UP:
			{
				printf("UP\n");
				break;
			}
			default:
			{
				printf("");
				break;
			}
			}
	}
}





void PrintSolution (PuzNode *GoalNode)
{
	PuzNode * TempNode;
 
	if (NULL == GoalNode)
	{
		printf("Error Solution");
	}
	else
	{
		printf("Going backwords:\n");
		TempNode = GoalNode;
		
		while (NULL != TempNode->m_ptrParent)
		{
			switch (TempNode->m_cMoveThatGotMeHere)
			{
			case DOWN:
			{
				printf("DOWN\n");
				break;
			}
			case LEFT:
			{
				printf("LEFT\n");
				break;
			}
			case RIGHT:
			{
				printf("RIGHT\n");
				break;
			}
			case UP:
			{
				printf("UP\n");
				break;
			}
		default:
			{
			}
		}
			TempNode = TempNode->m_ptrParent;
		}
	}
	return;
}



/************ DEBUG FUNCTIONS*/ 
void PrintNode(PuzNode *Node)
{
	int i = 0;
	printf("***************\n");
	for (i=0; i<5; i++)
	{
		printf("%02d-", (Node->m_strPuz[i]-ASCIISTART) );
	}
	printf("\n");
	for (i=5; i<10; i++)
	{
		printf("%02d-", (Node->m_strPuz[i]-ASCIISTART) );
	}
	printf("\n");
	for (i=10; i<15; i++)
	{
		printf("%02d-", (Node->m_strPuz[i]-ASCIISTART) );
	}
	printf("\n");
	for (i=15; i<20; i++)
	{
		printf("%02d-", (Node->m_strPuz[i]-ASCIISTART) );
	}
	printf("\n");
	for (i=20; i<25; i++)
	{
		printf("%02d-", (Node->m_strPuz[i]-ASCIISTART) );
	}
	printf("\n");
	printf("***************\n");
}

PuzNode* CreateBadNode(void)
{
   PuzNode *badNode = (PuzNode *)malloc(sizeof(PuzNode));

   badNode->m_iMovesToGetHere = 0;   
   badNode->m_cHeuristic = 2;
   badNode->m_cMoveThatGotMeHere = NOMOVE;
   badNode->m_ptrParent = NULL;
	
/* Insert characters into node representation */
   badNode->m_strPuz[1] = 'A';
   badNode->m_strPuz[0] = 'B';
   badNode->m_strPuz[2] = 'C';
   badNode->m_strPuz[3] = 'D';
   badNode->m_strPuz[4] = 'E';
   badNode->m_strPuz[5] = 'F';
   badNode->m_strPuz[6] = 'G';
   badNode->m_strPuz[7] = 'H';
   badNode->m_strPuz[8] = 'I';
   badNode->m_strPuz[9] = 'J';
   badNode->m_strPuz[10] = 'K';
   badNode->m_strPuz[11] = 'L';
   badNode->m_strPuz[12] = 'M';
   badNode->m_strPuz[13] = 'N';
   badNode->m_strPuz[14] = 'O';
   badNode->m_strPuz[15] = 'P';
   badNode->m_strPuz[16] = 'Q';
   badNode->m_strPuz[17] = 'R';
   badNode->m_strPuz[18] = 'S';
   badNode->m_strPuz[19] = 'T';
   badNode->m_strPuz[20] = 'U';
   badNode->m_strPuz[21] = 'V';
   badNode->m_strPuz[22] = 'W';
   badNode->m_strPuz[23] = 'X';
   badNode->m_strPuz[24] = '@';
   badNode->m_strPuz[25] = '\0'; /* NULL terminator for string */

   return badNode;
}	

/************ DEBUG FUNCTIONS */ 
